@@ -1,2 +1,3 @@
 ﻿import MetaRagTool.Apps.GradioApps
-print("Apps imported")
\ No newline at end of file
+
+# print("Apps imported")
\ No newline at end of file
