@@ -2,7 +2,7 @@
 import weave
 
 
-class JudgeLLM():
+class JudgeLLM:
     def __init__(self,model_name=Gemini.GEMINI_2_FLASH_LIGHT,RequestPerMinute_limit=30):
         sys_message = """You are an impartial judge evaluating answer quality.
         Your task is to determine if an answer correctly addresses a question based on the ground truth.
