@@ -2,22 +2,23 @@
 import faiss
 import numpy as np
 from hazm import Normalizer
-from langchain_text_splitters import RecursiveCharacterTextSplitter
-import matplotlib.pyplot as plt
-from MetaRagTool.RAG.DocumentStructs import MySentence, MyDocument, MyParagraph, MyChunk, ChunkingMethod
 
-from MetaRagTool.Utils.MyUtils import capped_sent_tokenize, token_len, reflect_vector, remove_duplicates
+from MetaRagTool.RAG.Chunkers import ChunkerFactory
+from MetaRagTool.RAG.DocumentStructs import MRSentence, MRDocument, MRParagraph, generate_document_structure
+from MetaRagTool.RAG.Chunkers import ChunkingMethod
+
+from MetaRagTool.Utils.MRUtils import capped_sent_tokenize, token_len, reflect_vector, remove_duplicates
 
 from MetaRagTool.LLM.LLMIdentity import LLMIdentity
-from MetaRagTool.Encoders.MyEncoder import MyEncoder
-from MetaRagTool.Encoders.MyReranker import MyReranker
+from MetaRagTool.Encoders.Encoder import Encoder
+from MetaRagTool.Encoders.Reranker import Reranker
 
 
 
 class MetaRAG:
 
 
-    def __init__(self, encoder_model: MyEncoder, llm: LLMIdentity, reranker_model: MyReranker = None,
+    def __init__(self, encoder_model: Encoder, llm: LLMIdentity=None, reranker_model: Reranker = None,
                  splitting_method=ChunkingMethod.SENTENCE_MERGER, chunk_size=90, chunk_overlap=3, max_sentence_len=-1,
                  use_neighbor_embeddings=False
                  , use_parentParagraph_embeddings=False, add_neighbor_chunks=False, add_neighbor_chunks_smart=False,
@@ -27,7 +28,7 @@ class MetaRAG:
                  rerank=False,
                  additional_top_k=5, include_reflections=False,
                  include_refractions=False, sent_merge_merged_chunks=True, log_chunking_report=True, weighted_bfs=False,
-                 embedding_steering_influence_factor=0.35,add_neighbor_chunks_k=2):
+                 embedding_steering_influence_factor=0.35, add_neighbor_chunks_k=2):
 
 
 
@@ -43,8 +44,8 @@ class MetaRAG:
         self.encoder_model = encoder_model
         self.chunk_overlap = chunk_overlap
         self.index = None
-        self.myDocuments = []
-        self.myChunks = []
+        self.DocumentsList = []
+        self.ChunksList = []
         self.use_neighbor_embeddings = use_neighbor_embeddings
         self.use_parentParagraph_embeddings = use_parentParagraph_embeddings
         self.add_neighbor_chunks = add_neighbor_chunks
@@ -66,6 +67,13 @@ class MetaRAG:
         self.time_to_encode_corpus = 0
         self.answer_mode='none'
 
+        self.chunker = None
+
+    def apply_best_config(self,text_has_proper_paragraphing=True):
+        from MetaRagTool.Utils import get_best_config
+        bestConfig = get_best_config(text_has_proper_paragraphing=True)
+        self.apply_config(bestConfig)
+
 
 
     def report(self):
@@ -119,17 +127,14 @@ class MetaRAG:
         self.weighted_bfs = ragConfig.weighted_bfs
         self.additional_top_k = ragConfig.additional_top_k
         self.include_reflections_k = ragConfig.include_reflections_k
-
-
-
-
-
+        self.rerank = ragConfig.rerank
+        self.reranker_model = ragConfig.reranker
 
     def _encode_and_index_dataset(self):
         if self.use_parentParagraph_embeddings:
             self._set_paragraph_embeddings()
 
-        chunks_to_encode = [chunk for chunk in self.myChunks if chunk.Embeddings is None]
+        chunks_to_encode = [chunk for chunk in self.ChunksList if chunk.Embeddings is None]
         print(f"Number of chunks to encode: {len(chunks_to_encode)}")
         chunks_texts = [self.normalizer.normalize(chunk.Text) if self.normalize_text else chunk.Text for chunk in
                         chunks_to_encode]
@@ -186,265 +191,62 @@ class MetaRAG:
         for i, paragraph in enumerate(paragraphs_to_encode):
             paragraph.Embeddings = paragraph_embeddings[i]
 
-    def _generate_document_structure(self, rawDocumentsTexts):
-        if self.max_sentence_len == -1:
-            self.max_sentence_len = self.chunk_size
-            print("Error: max_sentence_len is set to -1, which is invalid.")
-
-
-        # Generate the document structure
-        for documentText in rawDocumentsTexts:
-            if len(documentText) < 1:
-                continue
-
-            newDoc = MyDocument(documentText)
-            self.myDocuments.append(newDoc)
-
-            prevParagraph = None
-            for paragraphText in documentText.split('\n'):
-                if len(paragraphText) < 5:
-                    continue
-
-                newParagraph = MyParagraph(document=newDoc, text=paragraphText)
-
-                prevSentence = None
-                sentences = capped_sent_tokenize(paragraphText, self.max_sentence_len)
-                if len(sentences) < 1:
-                    continue
-
-                for sentenceText in sentences:
-
-                    newSentence = MySentence(document=newDoc, paragraph=newParagraph, text=sentenceText)
-                    newParagraph.AddSentence(newSentence)
-                    newDoc.AddSentence(newSentence)
-                    if prevSentence is not None:
-                        prevSentence.SetNext(newSentence)
-                        newSentence.SetPrev(prevSentence)
-
-                    prevSentence = newSentence
-
-                newDoc.AddParagraph(newParagraph)
-                if prevParagraph is not None:
-                    prevParagraph.SetNext(newParagraph)
-                    newParagraph.SetPrev(prevParagraph)
-
-                prevParagraph = newParagraph
-
-        self.allParagraphs = [paragraph for doc in self.myDocuments for paragraph in doc.MyParagraphs]
-        self.allSentences = [sentence for doc in self.myDocuments for sentence in doc.MySentences]
-        print(
-            f"Corpus structure: {len(self.myDocuments)} documents and {len(self.allParagraphs)} paragraphs and {len(self.allSentences)} sentences")
 
     def _chunkify(self):
 
-        recursive_text_splitter = RecursiveCharacterTextSplitter(
-            chunk_size=self.chunk_size,
-            chunk_overlap=self.chunk_overlap,
-            length_function=token_len,
-            is_separator_regex=False,
-        )
+        if self.chunker is None:
+            self.chunker = ChunkerFactory.create_chunker(splitting_method=self.splitting_method, chunksList=self.ChunksList,
+                                                         chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap, sent_merge_merged_chunks=self.sent_merge_merged_chunks)
 
-        for myDoc in self.myDocuments:
-            if myDoc.isChunked:
-                continue
+        self.chunker.chunk_documents(self.DocumentsList)
 
-            myDoc.isChunked = True
-            if self.splitting_method == ChunkingMethod.SENTENCE:
-                prevChunk = None
-                for sentence in myDoc.MySentences:
-                    newChunk = MyChunk(document=myDoc, text=sentence.Text)
-                    newChunk.AddParagraph(sentence.MyParagraph)
-                    newChunk.AddSentence(sentence)
-                    self.myChunks.append(newChunk)
-                    if prevChunk is not None:
-                        prevChunk.NextRelated = newChunk
-                        newChunk.PrevRelated = prevChunk
-                    prevChunk = newChunk
-
-            elif self.splitting_method == ChunkingMethod.PARAGRAPH:
-                prevChunk = None
-                for paragraph in myDoc.MyParagraphs:
-                    newChunk = MyChunk(document=myDoc, text=paragraph.Text)
-                    newChunk.AddParagraph(paragraph)
-                    for sentence in paragraph.MySentences:
-                        newChunk.AddSentence(sentence)
-
-                    self.myChunks.append(newChunk)
-                    if prevChunk is not None:
-                        prevChunk.NextRelated = newChunk
-                        newChunk.PrevRelated = prevChunk
-                    prevChunk = newChunk
+        if self.log_chunking_report:self.chunker.chunking_report()
 
 
-            elif self.splitting_method == ChunkingMethod.DOCUMENT:
-                newChunk = MyChunk(document=myDoc, text=myDoc.Text)
-                for paragraph in myDoc.MyParagraphs:
-                    newChunk.AddParagraph(paragraph)
-                    for sentence in paragraph.MySentences:
-                        newChunk.AddSentence(sentence)
-
-                self.myChunks.append(newChunk)
-
-
-
-            elif self.splitting_method == ChunkingMethod.RECURSIVE:
-                chunks = recursive_text_splitter.split_text(myDoc.Text)
-                for chunk in chunks:
-                    self.myChunks.append(MyChunk(document=myDoc, text=chunk))
-
-
-
-            elif self.splitting_method == ChunkingMethod.SENTENCE_MERGER:
-                prevChunk = None
-
-                for paragraph in myDoc.MyParagraphs:
-                    current_chunk_sentences = []
-                    currentChunkText = ""
-                    currentSentence: MySentence = paragraph.MySentences[0]
-                    while currentSentence is not None:
-                        nextText = currentSentence.Text + " "
-                        # If adding this sentence would exceed chunk_size, finalize current chunk first.
-                        if token_len(currentChunkText + nextText) > self.chunk_size and currentChunkText.strip():
-                            currentChunkText, current_chunk_sentences, prevChunk = self.__sent_merger_create_chunk(currentChunkText,
-                                                                                                    current_chunk_sentences,
-                                                                                                    myDoc, paragraph,
-                                                                                                    prevChunk)
-
-                        currentChunkText += nextText
-                        current_chunk_sentences.append(currentSentence)
-
-                        # If next sentence is None or we've just reached chunk size, finalize the chunk.
-                        if currentSentence.Next is None:
-                            currentChunkText, current_chunk_sentences, prevChunk = self.__sent_merger_create_chunk(currentChunkText,
-                                                                                                    current_chunk_sentences,
-                                                                                                    myDoc, paragraph,
-                                                                                                    prevChunk)
-
-                        currentSentence = currentSentence.Next
-
-            elif self.splitting_method == ChunkingMethod.SENTENCE_MERGER_CROSS_PARAGRAPH:
-                prevChunk = None
-                current_chunk_sentences = []
-                currentChunkText = ""
-
-                for idx, sentt in enumerate(myDoc.MySentences):
-                    # If adding this sentence exceeds chunk_size, create the chunk right away
-                    if currentChunkText and token_len(currentChunkText + sentt.Text) > self.chunk_size:
-                        newChunk = MyChunk(document=myDoc, text=currentChunkText)
-                        for s in current_chunk_sentences:
-                            newChunk.AddSentence(s)
-                            newChunk.AddParagraph(s.MyParagraph)
-                        self.myChunks.append(newChunk)
-                        if prevChunk is not None:
-                            prevChunk.NextRelated = newChunk
-                            newChunk.PrevRelated = prevChunk
-
-                        prevChunk = newChunk
-                        currentChunkText = ""
-                        current_chunk_sentences = []
-
-                    currentChunkText += sentt.Text + " "
-                    current_chunk_sentences.append(sentt)
-
-                    # If it's the last sentence, finalize the current chunk
-                    if idx == len(myDoc.MySentences) - 1:
-                        newChunk = MyChunk(document=myDoc, text=currentChunkText)
-                        for s in current_chunk_sentences:
-                            newChunk.AddSentence(s)
-                            newChunk.AddParagraph(s.MyParagraph)
-                        self.myChunks.append(newChunk)
-                        if prevChunk is not None:
-                            prevChunk.NextRelated = newChunk
-                            newChunk.PrevRelated = prevChunk
-
-
-        for chunk in self.myChunks:
-            chunk.Length = token_len(chunk.Text)
-
-
-        if self.log_chunking_report:self.chunking_report()
-
-    def __sent_merger_create_chunk(self, currentChunkText, current_chunk_sentences, myDoc, paragraph, prevChunk):
-        if self.sent_merge_merged_chunks and prevChunk is not None and token_len(prevChunk.Text + currentChunkText) < self.chunk_size:
-            prevChunk.Text += currentChunkText
-            for s in current_chunk_sentences:
-                prevChunk.AddSentence(s)
-            prevChunk.AddParagraph(paragraph)
-        else:
-            newChunk = MyChunk(document=myDoc, text=currentChunkText)
-            newChunk.AddParagraph(paragraph)
-            for s in current_chunk_sentences:
-                newChunk.AddSentence(s)
-            self.myChunks.append(newChunk)
-            if prevChunk is not None:
-                prevChunk.NextRelated = newChunk
-                newChunk.PrevRelated = prevChunk
-            prevChunk = newChunk
-        currentChunkText = ""
-        current_chunk_sentences = []
-
-
-
-        return currentChunkText, current_chunk_sentences, prevChunk
-
-    def chunking_report(self):
-        chunk_sizes = [token_len(chunk.Text) for chunk in self.myChunks]
-        total_chunks = len(chunk_sizes)
-        avg_chunk_size = sum(chunk_sizes) / total_chunks if total_chunks > 0 else 0
-        max_chunk_size = max(chunk_sizes) if total_chunks > 0 else 0
-        min_chunk_size = min(chunk_sizes) if total_chunks > 0 else 0
-        median_chunk_size = np.median(chunk_sizes) if total_chunks > 0 else 0
-        std_chunk_size = np.std(chunk_sizes) if total_chunks > 0 else 0
-        # chunk_size_distribution = np.histogram(chunk_sizes, bins='auto') if total_chunks > 0 else ([], [])
-
-        print(f"Chunking Report:")
-        print(f"Total number of chunks: {total_chunks}")
-        print(f"Average chunk size: {avg_chunk_size:.2f} tokens")
-        print(f"Maximum chunk size: {max_chunk_size} tokens")
-        print(f"Minimum chunk size: {min_chunk_size} tokens")
-        print(f"Median chunk size: {median_chunk_size:.2f} tokens")
-        print(f"Standard deviation of chunk size: {std_chunk_size:.2f} tokens")
-
-        plt.clf()
-        plt.hist(chunk_sizes, bins='auto', edgecolor='black')
-        plt.title('Chunk Size Distribution')
-        plt.xlabel('Chunk Size (tokens)')
-        plt.ylabel('Frequency')
-        plt.show()
-
-        return plt
 
 
     def add_corpus(self, raw_documents_text: list,encode=True):
-        self._generate_document_structure(raw_documents_text)
+        self.allParagraphs,self.allSentences=generate_document_structure(rawDocumentsTexts=raw_documents_text,DocumentsList=self.DocumentsList,
+                                    max_sentence_len=self.max_sentence_len if self.max_sentence_len != -1 else self.chunk_size)
+
+        print(f"Corpus structure: {len(self.DocumentsList)} documents and {len(self.allParagraphs)} paragraphs and {len(self.allSentences)} sentences")
+
+
         self._chunkify()
         if encode:self._encode_and_index_dataset()
 
-
-    def retrieve_core(self, query_embedding, top_k):
+    def _retrieve_core(self, query_embedding, top_k):
         if top_k < 1:
             print("Error: top_k is set to 0, which is invalid.")
             return []
 
         query_embedding = np.array(query_embedding).reshape(1, -1)
         distances, indices = self.index.search(query_embedding, top_k)
-        retrieved_chunks = [self.myChunks[i] for i in indices[0]]
+        retrieved_chunks = [self.ChunksList[i] for i in indices[0]]
         return retrieved_chunks
 
 
-    def retrieve(self, query, top_k=20, force_basics=False, sum_of_retrieved_token_length_limit=-1):
+    def _check_feature_compatibility(self):
+        if self.splitting_method == ChunkingMethod.RECURSIVE and (self.replace_retrieved_chunks_with_parent_paragraph or self.add_neighbor_chunks or self.add_neighbor_chunks_smart):
+            print("Error: Recursive splitting method is not compatible with replace_retrieved_chunks_with_parent_paragraph. Please set it to False.")
+            self.replace_retrieved_chunks_with_parent_paragraph = False
+            self.add_neighbor_chunks_smart = False
+            self.add_neighbor_chunks = False
 
-        if sum_of_retrieved_token_length_limit>0:
-            # fixing top_k
-            top_k = sum_of_retrieved_token_length_limit // self.chunk_size
-            if self.replace_retrieved_chunks_with_parent_paragraph:
-                top_k = top_k // 2
-            if self.include_reflections or self.include_refractions:
-                top_k = top_k // 2
-            if self.add_neighbor_chunks or self.add_neighbor_chunks_smart:
-                top_k = top_k // 2
-            if self.breath_first_retrival or self.depth_first_retrival or self.weighted_bfs:
-                top_k = top_k // 2
+    def retrieve(self, query, top_k=20, force_basics=False):
+        # if sum_of_retrieved_token_length_limit>0:
+        #     # fixing top_k
+        #     top_k = sum_of_retrieved_token_length_limit // self.chunk_size
+        #     if self.replace_retrieved_chunks_with_parent_paragraph:
+        #         top_k = top_k // 2
+        #     if self.include_reflections or self.include_refractions:
+        #         top_k = top_k // 2
+        #     if self.add_neighbor_chunks or self.add_neighbor_chunks_smart:
+        #         top_k = top_k // 2
+        #     if self.breath_first_retrival or self.depth_first_retrival or self.weighted_bfs:
+        #         top_k = top_k // 2
+
+        self._check_feature_compatibility()
 
 
         if top_k<1:
@@ -457,28 +259,32 @@ class MetaRAG:
 
         query_embedding = self.encoder_model.encode([query])
 
-        initial_k = top_k
+        retrieved_chunks = self._retrieve_core(query_embedding, top_k*5 if self.rerank else top_k)
+
+
         if self.rerank:
-            top_k = 2 * top_k
+            if self.reranker_model is None: print("Error: Reranker model is not set. Please set the Reranker model before using it or set rerank=False")
+
+            retrieved_chunks=self.reranker_model.apply_rerank_MRChunks(query=query, chunks=retrieved_chunks)
+            retrieved_chunks = retrieved_chunks[: top_k]
 
-        retrieved_chunks = self.retrieve_core(query_embedding, initial_k)
 
         if self.breath_first_retrival and not force_basics:
             additional_chunks = []
-            top_k_retrieved_chunks = retrieved_chunks[:initial_k // 5]
+            top_k_retrieved_chunks = retrieved_chunks[:top_k // 5]
             for chunk in top_k_retrieved_chunks:
-                additional_chunks.extend(self.retrieve_core(chunk.Embeddings, self.additional_top_k))
+                additional_chunks.extend(self._retrieve_core(chunk.Embeddings, self.additional_top_k))
             retrieved_chunks.extend(additional_chunks)
             retrieved_chunks = remove_duplicates(retrieved_chunks)
 
 
-        elif self.depth_first_retrival and not force_basics and initial_k > 1:
+        elif self.depth_first_retrival and not force_basics and top_k > 1:
             # repeat retrieval for the first chunk with the original top_k value, depth_first_retrival_k times, each time with the retrieved chunks (like going deeper only in one branch of the tree)
             additional_chunks = []
             # Replace the empty section with:
             current_chunk = retrieved_chunks[0]  # Start with first retrieved chunk
             for _ in range(2):
-                level_chunks = self.retrieve_core(current_chunk.Embeddings, initial_k // 2)
+                level_chunks = self._retrieve_core(current_chunk.Embeddings, top_k // 2)
                 additional_chunks.extend(level_chunks)
                 current_chunk = level_chunks[0]  # Go deeper with first chunk
 
@@ -488,24 +294,15 @@ class MetaRAG:
         elif self.weighted_bfs and not force_basics:
             additional_chunks = []
             for i, chunk in enumerate(retrieved_chunks):
-                chunk_top_k = initial_k // (2 ** (i + 1))
+                chunk_top_k = top_k // (2 ** (i + 1))
                 if chunk_top_k < 1:
                     break
-                additional_chunks.extend(self.retrieve_core(chunk.Embeddings, chunk_top_k))
+                additional_chunks.extend(self._retrieve_core(chunk.Embeddings, chunk_top_k))
 
             retrieved_chunks.extend(additional_chunks)
             retrieved_chunks = remove_duplicates(retrieved_chunks)
 
 
-
-
-        if self.rerank:
-            scores = self.reranker_model.rerank(query, retrieved_chunks)
-            # scores is a list of floats with the same length as retrieved_chunks, and higher score means better chunk
-
-            retrieved_chunks = [chunk for _, chunk in sorted(zip(scores, retrieved_chunks), reverse=True)]
-            retrieved_chunks = retrieved_chunks[:top_k]
-
         if self.include_reflections:
             # for each retrieved chunk, get its reflection based on query and retrived only 1 more chunks with the new vector
             additional_chunks = []
@@ -513,7 +310,7 @@ class MetaRAG:
                 reflection = reflect_vector(query_embedding[0], chunk.Embeddings)
                 reflection = np.array([reflection], dtype=np.float32)
                 faiss.normalize_L2(reflection)
-                additional_chunks.extend(self.retrieve_core(reflection, 1))
+                additional_chunks.extend(self._retrieve_core(reflection, 1))
 
             retrieved_chunks.extend(additional_chunks)
             retrieved_chunks = remove_duplicates(retrieved_chunks)
@@ -526,7 +323,7 @@ class MetaRAG:
                 newV = np.array([newV], dtype=np.float32)
                 faiss.normalize_L2(newV)
 
-                closest_chunk = self.retrieve_core(newV, 1)
+                closest_chunk = self._retrieve_core(newV, 1)
                 additional_chunks.extend(closest_chunk)
 
             retrieved_chunks.extend(additional_chunks)
@@ -590,26 +387,26 @@ class MetaRAG:
         else:
             retrieved_chunks_texts = [chunk.Text for chunk in retrieved_chunks]
 
-        if sum_of_retrieved_token_length_limit > 0:
-            total_tokens = 0
-            filtered_chunks = []
-            for chunk in retrieved_chunks_texts:
-                c_len = token_len(chunk)
-                filtered_chunks.append(chunk)
-                if total_tokens + c_len > sum_of_retrieved_token_length_limit:
-                    break
-                total_tokens += c_len
-            retrieved_chunks_texts = filtered_chunks
+        # if sum_of_retrieved_token_length_limit > 0:
+        #     total_tokens = 0
+        #     filtered_chunks = []
+        #     for chunk in retrieved_chunks_texts:
+        #         c_len = token_len(chunk)
+        #         filtered_chunks.append(chunk)
+        #         if total_tokens + c_len > sum_of_retrieved_token_length_limit:
+        #             break
+        #         total_tokens += c_len
+        #     retrieved_chunks_texts = filtered_chunks
 
 
         return retrieved_chunks_texts
 
-    def __ask_notTool(self, query, top_k=1, include_prompt=True):
+    def _ask_notTool(self, query, top_k=30, include_prompt=True):
         if self.answer_mode=='none':
             self.answer_mode = 'classic'
         elif self.answer_mode=='tool':
             print("changing mode to classic, chat history is cleared")
-            self.llm.messages_history=[]
+            self.clear_history()
 
         # encode query -> retrieve top_k chunks -> generate answer
 
@@ -625,31 +422,29 @@ class MetaRAG:
 
         return answer
 
-    def ask(self, query, top_k=1, include_prompt=True,useTool=False):
+    def ask(self, query, top_k=30, include_prompt=False,useTool=False):
+        if self.llm is None:
+            print("Error: LLM is not set. Please set the LLM before asking a question. or use retrieve()")
+            return ""
+
         if useTool:
-            return self.__ask_tool(query)
+            return self._ask_tool(query)
         else:
-            return self.__ask_notTool(query, top_k=top_k, include_prompt=include_prompt)
+            return self._ask_notTool(query, top_k=top_k, include_prompt=include_prompt)
 
-
-
-
-    def __ask_tool(self, query):
+    def _ask_tool(self, query):
         print("using tool")
         if self.answer_mode=='none':
             self.answer_mode = 'tool'
         elif self.answer_mode=='classic':
             print("changing mode to tool, chat history is cleared")
-            self.llm.messages_history=[]
+            self.clear_history()
 
         response = self.llm.generate(prompt=query,tool_function= self.retrieve_interface)
         return response
 
-
     def clear_history(self):
-        self.llm.messages_history = []
-
-
+        self.llm.reset_history()
 
     def retrieve_interface(self, query:str, top_k:int) -> str:
         """retrieves k chunks based on query.
@@ -661,7 +456,7 @@ class MetaRAG:
         """
         retrieved_chunks=self.retrieve(query=query, top_k=top_k)
 
-        retrieved_chunks_text=self.LLMIdentity.merge_chunks(retrieved_chunks=retrieved_chunks)
+        retrieved_chunks_text=LLMIdentity.merge_chunks(retrieved_chunks=retrieved_chunks)
 
         retrieved_chunks_text += f"\n if the retrieved chunks didn't contain the information you needed, try to increase the top_k value or use different query. call the tool again right now with function_call before talking to the user."
 
